import java.util.*;

class Resolver {
    List<Clause> knowledgeBase = new ArrayList<>();

    void updateKB(List<Clause> KB) {this.knowledgeBase = KB;}

    void addClause(Clause clause) {
        knowledgeBase.add(clause);
    }

    boolean resolve(Clause goal) {
		Queue<Clause> toResolve = new LinkedList<>(knowledgeBase);
		Set<Clause> seenClauses = new HashSet<>(knowledgeBase); // Track processed clauses
		toResolve.add(goal);
	
		while (!toResolve.isEmpty()) {
			Clause clause = toResolve.poll();
	
			for (Clause kbClause : knowledgeBase) {
				System.out.println("Resolving " + clause + " with " + kbClause);
				Clause resolvedClause = resolveClauses(clause, kbClause);
	
				if (resolvedClause != null && resolvedClause.predicates.isEmpty()) {
					// Contradiction found: Empty clause
					return true;
				}
	
				if (resolvedClause != null && !seenClauses.contains(resolvedClause)) {
					// Add only new clauses
					seenClauses.add(resolvedClause);
					toResolve.add(resolvedClause);
				}
			}
		}
		return false; // No contradiction, goal is not derivable
	}


    private Clause resolveClauses(Clause c1, Clause c2) {
        for (Predicate p1 : c1.predicates) {
            for (Predicate p2 : c2.predicates) {
                if (p1.isNegation(p2)) {
                    Unifier unifier = new Unifier();
                    boolean canUnify = true;
                    for (int i = 0; i < p1.terms.size(); i++) {
                        if (!unifier.unify(p1.terms.get(i), p2.terms.get(i))) {
                            canUnify = false;
                            break;
                        }
                    }
                    if (canUnify) {
                        Set<Predicate> newPredicates = new HashSet<>();
                        newPredicates.addAll(c1.predicates);
                        newPredicates.addAll(c2.predicates);
                        newPredicates.remove(p1);
                        newPredicates.remove(p2);
                        return new Clause(new ArrayList<>(newPredicates));
                    }
                }
            }
        }
        return null;
    }
}